class DatasetTypes:
    IC_MULTILABEL = 'classification_multilabel'
    IC_MULTICLASS = 'classification_multiclass'
    OD = 'object_detection'
    MULTITASK = 'multitask'
    IMCAP = 'image_caption'

    VALID_TYPES = [IC_MULTILABEL, IC_MULTICLASS, OD, MULTITASK, IMCAP]

    @staticmethod
    def is_classification(dataset_type):
        return dataset_type.startswith('classification')


class Usages:
    TRAIN_PURPOSE = 'train'
    VAL_PURPOSE = 'val'
    TEST_PURPOSE = 'test'


class Formats:
    IRIS = 'iris'
    COCO = 'coco'


class BBoxFormat:
    LTRB = 'ltrb'
    LTWH = 'ltwh'

    VALID_TYPES = [LTRB, LTWH]

    @staticmethod
    def validate(bbox_format):
        assert bbox_format in BBoxFormat.VALID_TYPES, f'Invalid bbox_format: {bbox_format}. Must be {BBoxFormat.VALID_TYPES}'



"""
Converts tsv format to coco format
ic: img_id  class_id    img_data_base64
od: img_id  [{"class": class_name, "rect": [L, T, R, B], "diff": 0}, ...] img_data_base64
"""

import logging
import json
import os

from .utils import verify_and_correct_box_or_none, guess_encoding, decode64_to_pil, zip_folder, TSV_FORMAT_LTRB, TSV_FORMAT_LTWH_NORM

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def create_arg_parser():
    import argparse

    parser = argparse.ArgumentParser(description='Convert tsv data to coco format.')
    parser.add_argument('-t', '--tsvs', required=True, nargs='+', help='Tsv files to convert.')
    parser.add_argument('-c', '--task', required=True, type=str, help='type of tasks.', choices=['ic', 'od'])
    parser.add_argument('-l', '--labelmap', type=str, default='labelmap.txt')
    parser.add_argument('-f', '--format', type=str, default=TSV_FORMAT_LTRB, choices=[TSV_FORMAT_LTRB, TSV_FORMAT_LTWH_NORM])
    parser.add_argument('-d', '--difficulty', type=bool, default=False, help='Include difficulty boxes or not.')
    parser.add_argument('-z', '--zip', type=bool, default=False, help='Zip the image and label folders or not.')
    parser.add_argument('-o', '--output_folder', type=str, required=True, help='Output folder for images.')

    return parser


def main():
    args = create_arg_parser().parse_args()
    logger.info(args.__dict__)
    image_folder_name = args.output_folder

    if not os.path.exists(image_folder_name):
        os.mkdir(image_folder_name)

    labelmap_exists = os.path.exists(args.labelmap)
    if labelmap_exists:
        logger.info(f'Labelmap {args.labelmap} exists.')
        with open(args.labelmap, 'r') as labelmap_in:
            categories = [x.strip() for x in labelmap_in.readlines()]
            if not categories:
                raise Exception(f'Empty labelmap {args.labelmap}')
            label_name_to_idx = {x: i for i, x in enumerate(categories)}
            categories = [{'name': x, 'id': i + 1} for i, x in enumerate(categories)]
    else:
        assert args.task != 'ic', 'labelmap must exist for ic.'
        logger.info(f'Labelmap {args.labelmap} does not exist, created on the fly.')
        label_name_to_idx = dict()
        categories = None

    for tsv_file_name in args.tsvs:
        line_idx = 1
        images = []
        annotations = []
        with open(tsv_file_name, 'r', encoding=guess_encoding(tsv_file_name)) as file_in:
            logger.info(f'Processing {tsv_file_name}.')
            for img_info in file_in:
                img_id, labels, img_b64 = img_info.split('\t')

                img = decode64_to_pil(img_b64)
                w, h = img.size

                # image data => image file
                img_file_name = img_id.replace('/', '_') + '.' + img.format
                img.save(os.path.join(image_folder_name, img_file_name), img.format)

                image_path = f'{image_folder_name}.zip@{img_file_name}'

                img_info_dict = {'id': line_idx, 'width': w, 'height': h, 'file_name': image_path}
                images.append(img_info_dict)

                # image info => index file

                lp = f'File: {tsv_file_name}, Line {line_idx}: '

                if args.task == 'ic':
                    for label_idx in labels.split(','):
                        annotations.append({'id': len(annotations) + 1, 'category_id': int(label_idx) + 1, 'image_id': line_idx})
                else:
                    # labels => files
                    labels = json.loads(labels)
                    for label in labels:
                        difficulty = label['diff'] if 'diff' in label else 0
                        if difficulty > 0 and not args.difficulty:
                            continue

                        if label['class'] not in label_name_to_idx:
                            if labelmap_exists:
                                raise Exception(f'{lp}Illegal class {label["class"]}, not in provided labelmap.')
                            else:
                                label_name_to_idx[label['class']] = len(label_name_to_idx)

                        label_idx = label_name_to_idx[label['class']]
                        box = verify_and_correct_box_or_none(lp, label['rect'], args.format, w, h)
                        if box is None:
                            continue

                        annotations.append({'id': len(annotations) + 1, 'category_id': label_idx + 1, 'image_id': line_idx, 'bbox': [box[0], box[1], box[2] - box[0], box[3] - box[1]]})

                line_idx += 1
                if line_idx % 2000 == 0:
                    logger.info(f'Processed {line_idx} images.')

        categories = categories or [{'id': idx + 1, 'name': name} for name, idx in label_name_to_idx.items()]
        coco_file_name = os.path.splitext(os.path.basename(tsv_file_name))[0] + '.json'
        with open(coco_file_name, 'w') as coco_out:
            coco_out.write(json.dumps({'images': images, 'annotations': annotations, 'categories': categories}, indent=2))

    if not labelmap_exists:
        logger.info(f'Write labelmap to {args.labelmap}')
        with open(args.labelmap, 'w') as labelmap_out:
            idx_to_labels = {label_name_to_idx[key]: key for key in label_name_to_idx}
            for i in range(len(idx_to_labels)):
                labelmap_out.write(idx_to_labels[i] + '\n')

    if args.zip:
        logger.info(f'Zip folder "{image_folder_name}".')
        zip_folder(image_folder_name)


if __name__ == '__main__':
    main()



import collections
import copy
import json
import logging
import os
import random
from typing import List, Dict
from urllib import parse as urlparse

from .constants import DatasetTypes, Formats, BBoxFormat
from .dataset_info import MultiTaskDatasetInfo
from .util import is_url, FileReader

logger = logging.getLogger(__name__)


def purge_line(line):
    if not isinstance(line, str):
        line = line.decode('utf-8')

    return line.strip()


def _purge_path(path):
    assert path is not None

    return path.replace('\\', '/')


def _construct_full_path_generator(dirs: List[str]):
    """
    generate a function that appends dirs to a provided path, if dirs is empty, just return the path
    Args:
        dirs (str): dirs to be appended to a given path. None or empty str in dirs will be filtered.

    Returns:
        full_path_func: a func that appends dirs to a given path

    """
    dirs = [x for x in dirs if x]

    if dirs:
        def full_path_func(path):
            to_join = [x for x in dirs + [path] if x]
            return _purge_path(os.path.join(*to_join))
    else:
        full_path_func = _purge_path

    return full_path_func


def _add_path_to_sas(sas, path_or_dir):
    assert sas
    if not path_or_dir:
        return sas

    parts = urlparse.urlparse(sas)
    path = _purge_path(os.path.join(parts[2], path_or_dir))
    path = path.replace('.zip@', '/')  # cannot read from zip file with path targeting a url
    url = urlparse.urlunparse((parts[0], parts[1], path, parts[3], parts[4], parts[5]))
    return url


def _construct_full_sas_generator(container_sas: str):
    if not container_sas:
        return _purge_path

    def func(file_path):
        return _add_path_to_sas(container_sas, file_path)

    return func


def _construct_full_sas_or_path_generator(container_sas_or_root_dir, prefix_dir=None):
    if container_sas_or_root_dir and is_url(container_sas_or_root_dir):
        return lambda path: _construct_full_sas_generator(container_sas_or_root_dir)(_construct_full_path_generator([prefix_dir])(path))
    else:
        return lambda path: _construct_full_path_generator([container_sas_or_root_dir, prefix_dir])(path)


class ImageDataManifest:
    """
    Encapsulates the information and annotations of an image.

    img_path could be 1. a local path 2. a local path in a non-compressed zip file (`c:\a.zip@1.jpg`) or 3. a url.
    """

    def __init__(self, id, img_path, width, height, labels):
        """
        Args:
            id (int or str): image id
            img_path (str): path to image
            width (int): image width
            height (int): image height
            labels (list or dict):
                classification: [c_id] for multiclass, [c_id1, c_id2, ...] for multilabel;
                detection: [[c_id, left, top, right, bottom], ...];
                image caption: [caption1, caption2, ...];
                multitask: dict[task, labels]
        """
        self.id = id
        self.img_path = img_path
        self.width = width
        self.height = height
        self.labels = labels


class DatasetManifest:
    """
    Encapsulates every information about a dataset including labelmap, images (width, height, path to image), and annotations. Information about each image is encapsulated in ImageDataManifest.
    """

    def __init__(self, images: List[ImageDataManifest], labelmap, data_type):
        """

        Args:
            images (list): image manifest
            labelmap (list or dict): labels, or labels by task name
            data_type (str or dict) : data type, or data type by task name

        """
        assert data_type != DatasetTypes.MULTITASK, 'For multitask, data_type should be a dict mapping task name to concrete data type.'

        if isinstance(labelmap, dict):
            assert isinstance(data_type, dict), 'labelmap being a dict indicating this is a multitask dataset, however the data_type is not a dict.'
            assert labelmap.keys() == data_type.keys(), f'mismatched task names in labelmap and task_type: {labelmap.keys()} vs {data_type.keys()}'

        self.images = images
        self.labelmap = labelmap
        self.data_type = data_type

        self._task_names = sorted(labelmap.keys()) if self.is_multitask else None

    @staticmethod
    def create_dataset_manifest(dataset_info, usage: str, container_sas_or_root_dir: str = None):

        if dataset_info.data_format == Formats.IRIS:
            return IrisManifestAdaptor.create_dataset_manifest(dataset_info, usage, container_sas_or_root_dir)
        if dataset_info.data_format == Formats.COCO:
            get_full_sas_or_path = _construct_full_sas_or_path_generator(container_sas_or_root_dir, dataset_info.root_folder)('')
            if dataset_info.type == DatasetTypes.MULTITASK:
                coco_file_by_task = {k: sub_taskinfo.index_files.get(usage) for k, sub_taskinfo in dataset_info.sub_task_infos.items()}
                data_type_by_task = {k: sub_taskinfo.type for k, sub_taskinfo in dataset_info.sub_task_infos.items()}
                return CocoManifestAdaptor.create_dataset_manifest(coco_file_by_task, data_type_by_task, get_full_sas_or_path)

            return CocoManifestAdaptor.create_dataset_manifest(dataset_info.index_files.get(usage), dataset_info.type, get_full_sas_or_path)

        raise RuntimeError(f'{dataset_info.data_format} not supported yet.')

    @property
    def is_multitask(self):
        return isinstance(self.data_type, dict)

    def __len__(self):
        return len(self.images)

    def _add_label_count(self, labels, n_images_by_class: list):
        if self.is_multitask:
            for task_name, task_labels in labels.items():
                for label in task_labels:
                    n_images_by_class[self._get_cid(label, task_name)] += 1
        else:
            for label in labels:
                n_images_by_class[self._get_cid(label)] += 1

    def _get_label_count(self, labels, n_images_by_class: list):
        if self.is_multitask:
            return [n_images_by_class[self._get_cid(label, task_name)] for task_name, task_labels in labels.items() for label in task_labels]
        else:
            return [n_images_by_class[self._get_cid(label)] for label in labels]

    def _get_cid(self, label, task_name=None):
        if task_name:  # multitask
            cnt = 0
            for t_name in self._task_names:
                if t_name == task_name:
                    break
                cnt += len(self.labelmap[t_name])

            return cnt + self._get_cid(label)
        elif isinstance(label, int):  # classification
            return label
        elif isinstance(label, list):  # detection
            return label[0]
        else:
            raise RuntimeError(f'unknown type of label: {type(label)}')

    def _is_negative(self, labels):
        n_labels = len(labels) if not self.is_multitask else sum([len(x) for x in labels.values()])
        return n_labels == 0

    def generate_coco_annotations(self):
        """
        Generate coco annotations, working for single task classification, detection and caption only

        Returns:
            A dict of annotation data ready for coco json dump

        """

        images = []
        for i, x in enumerate(self.images):
            image = {'id': i + 1, 'file_name': x.img_path}
            if x.width:
                image['width'] = x.width
            if x.height:
                image['height'] = x.height
            images.append(image)

        annotations = []
        for img_id, img in enumerate(self.images):
            for ann in img.labels:
                coco_ann = {
                    'id': len(annotations) + 1,
                    'image_id': img_id + 1,
                }

                if DatasetTypes.is_classification(self.data_type):
                    coco_ann['category_id'] = ann + 1
                elif self.data_type == DatasetTypes.OD:
                    coco_ann['category_id'] = ann[0] + 1
                    coco_ann['bbox'] = [ann[1], ann[2], ann[3] - ann[1], ann[4] - ann[2]]
                elif self.data_type == DatasetTypes.IMCAP:
                    coco_ann['caption'] = ann
                else:
                    raise ValueError(f'Unsupported data type {self.data_type}')

                annotations.append(coco_ann)

        coco_dict = {'images': images, 'annotations': annotations}
        if self.data_type != DatasetTypes.IMCAP:
            coco_dict['categories'] = [{'id': i + 1, 'name': x} for i, x in enumerate(self.labelmap)]

        return coco_dict

    def train_val_split(self, train_ratio, random_seed=0):
        """
        Split the dataset into train and val set, with train set ratio being train_ratio.
        For multiclass dataset, the split ratio will be close to provided train_ratio, while for multilabel dataset, it is not guaranteed
        Multitask dataset and detection dataset are treated the same with multilabel dataset.
        Args:
            train_ratio(float): rough train set ratio, from 0 to 1
            random_seed: random seed

        Returns:
            train_manifest, val_manifest
        """
        if int(len(self.images) * train_ratio) == 0:
            return DatasetManifest([], self.labelmap, self.data_type), DatasetManifest(self.images, self.labelmap, self.data_type)

        if int(len(self.images) * train_ratio) == len(self.images):
            return DatasetManifest(self.images, self.labelmap, self.data_type), DatasetManifest([], self.labelmap, self.data_type)

        rng = random.Random(random_seed)
        images = list(self.images)
        rng.shuffle(images)

        train_imgs = []
        val_imgs = []
        n_train_imgs_by_class = [0] * len(self.labelmap) if not self.is_multitask else [0] * sum([len(x) for x in self.labelmap.values()])
        n_val_imgs_by_class = [0] * len(self.labelmap) if not self.is_multitask else [0] * sum([len(x) for x in self.labelmap.values()])
        test_train_ratio = (1 - train_ratio) / train_ratio
        n_train_neg = 0
        n_val_neg = 0
        for image in images:
            if self._is_negative(image.labels):
                if n_train_neg == 0 or n_val_neg / n_train_neg >= test_train_ratio:
                    n_train_neg += 1
                    train_imgs.append(image)
                else:
                    n_val_neg += 1
                    val_imgs.append(image)

                continue

            train_cnt = self._get_label_count(image.labels, n_train_imgs_by_class)
            val_cnt = self._get_label_count(image.labels, n_val_imgs_by_class)
            train_cnt_sum = sum(train_cnt) * test_train_ratio
            train_cnt_min = min(train_cnt) * test_train_ratio
            val_cnt_sum = sum(val_cnt)
            val_cnt_min = min(val_cnt)
            if val_cnt_min < train_cnt_min or (val_cnt_min == train_cnt_min and val_cnt_sum < train_cnt_sum):
                val_imgs.append(image)
                self._add_label_count(image.labels, n_val_imgs_by_class)
            else:
                train_imgs.append(image)
                self._add_label_count(image.labels, n_train_imgs_by_class)

        return DatasetManifest(train_imgs, self.labelmap, self.data_type), DatasetManifest(val_imgs, self.labelmap, self.data_type)

    def sample_categories(self, category_indices: List):
        """
        Sample a new dataset of selected categories. Works for single IC and OD dataset only.
        Args:
            category_indices: indices of the selected categories

        Returns:
            a sampled dataset with selected categories

        """

        assert self.data_type in [DatasetTypes.IC_MULTICLASS, DatasetTypes.IC_MULTILABEL, DatasetTypes.OD]
        assert category_indices
        assert max(category_indices) < len(self.labelmap)

        category_id_remap = {o_cid: n_cid for n_cid, o_cid in enumerate(category_indices)}
        new_labelmap = [self.labelmap[x] for x in category_indices]
        new_images = []
        for img in self.images:
            new_img = copy.deepcopy(img)
            if DatasetTypes.is_classification(self.data_type):
                new_img.labels = [category_id_remap[x] for x in new_img.labels if x in category_id_remap]
            else:
                new_img.labels = [[category_id_remap[x[0]], x[1], x[2], x[3], x[4]] for x in img.labels if x[0] in category_id_remap]

            new_images.append(new_img)
        return DatasetManifest(new_images, new_labelmap, self.data_type)

    def sample_subset(self, num_samples, with_replacement=False, random_seed=0):
        """
        Sample a subset of num_samples images. When with_replacement is False and num_samples is larger than the dataset, the whole dataset will be returned
        Args:
            num_samples (int): number of images to be sampled
            with_replacement (bool): with replacement or not
            random_seed (int): random seed

        Returns:
            a sampled dataset
        """

        rnd = random.Random(random_seed)
        if not with_replacement:
            if num_samples >= len(self.images):
                sampled_images = self.images
            else:
                sampled_images = rnd.sample(self.images, num_samples)
        else:
            sampled_images = [rnd.choice(self.images) for _ in range(num_samples)]

        return DatasetManifest(sampled_images, self.labelmap, self.data_type)

    def sample_few_shot_subset(self, num_samples_per_class, random_seed=0):
        """
        Sample a few-shot dataset, with the number of images per class below num_samples_per_class.
        For multiclass dataset, this is always possible, while for multilabel dataset, it is not guaranteed
        Multitask dataset and detection dataset are treated the same with multilabel dataset.

        This method tries to get balanced results.

        Note that negative images will be added to the subset up to num_samples_per_class.

        Args:
            num_samples_per_class: rough number samples per class to sample
            random_seed: random seed

        Returns:
            a sampled few-shot subset
        """

        assert num_samples_per_class > 0

        sampled_images = []
        rng = random.Random(random_seed)
        images = list(self.images)
        rng.shuffle(images)
        n_imgs_by_class = [0] * len(self.labelmap) if not self.is_multitask else [0] * sum([len(x) for x in self.labelmap.values()])
        neg_img_cnt = 0
        for image in images:
            if self._is_negative(image.labels):
                if neg_img_cnt < num_samples_per_class:
                    neg_img_cnt += 1
                    sampled_images.append(image)
                continue

            img_label_cnt = self._get_label_count(image.labels, n_imgs_by_class)

            if min(img_label_cnt) >= num_samples_per_class:
                continue

            if min(img_label_cnt) <= num_samples_per_class / 2 or max(img_label_cnt) <= 1.5 * num_samples_per_class:
                sampled_images.append(image)
                self._add_label_count(image.labels, n_imgs_by_class)

            if min(n_imgs_by_class) >= num_samples_per_class:
                break

        return DatasetManifest(sampled_images, self.labelmap, self.data_type)

    def sample_subset_by_ratio(self, sampling_ratio):
        """
        Sample a dataset so that each labels appears by at least the given sampling_ratio. In case of multiclass dataset, the number of sampled images will be N * sampling_ratio.
        For multilabel or object detection datasets, the total number of images will be bigger than that.

        Args:
            sampling_ratio (float): sampling ratio. must be 0 < x < 1.

        Returns:
            A sampled dataset (DatasetManifest)
        """
        assert 0 < sampling_ratio < 1

        if self.is_multitask:
            labels = [[self._get_cid(c, t) for t, t_labels in image.labels.items() for c in t_labels] for image in self.images]
        else:
            labels = [[self._get_cid(c) for c in image.labels] for image in self.images]

        # Create a dict {label_id: [image_id, ...], ...}
        # Note that image_id can be included multiple times if the dataset is multilabel, objectdetection, or multitask.
        label_image_map = collections.defaultdict(list)
        for i, image_labels in enumerate(labels):
            if not image_labels:
                label_image_map[-1].append(i)
            for label in image_labels:
                label_image_map[label].append(i)

        # From each lists, sample max(1, N * ratio) images.
        sampled_image_ids = set()
        for image_ids in label_image_map.values():
            sampled_image_ids |= set(random.sample(image_ids, max(1, int(len(image_ids) * sampling_ratio))))

        sampled_images = [self.images[i] for i in sampled_image_ids]
        return DatasetManifest(sampled_images, self.labelmap, self.data_type)

    def sample_few_shots_subset_greedy(self, num_min_samples_per_class, random_seed=0):
        """Greedy few-shots sampling method.
        Randomly pick images from the original datasets until all classes have at least {num_min_images_per_class} tags/boxes.

        Note that images without any tag/box will be ignored. All images in the subset will have at least one tag/box.

        Args:
            num_min_samples_per_class (int): The minimum number of samples per class.
            random_seed (int): Random seed to use.

        Returns:
            A samped dataset (DatasetManifest)

        Raises:
            RuntimeError if it couldn't find num_min_samples_per_class samples for all classes
        """
        assert num_min_samples_per_class > 0
        images = list(self.images)
        rng = random.Random(random_seed)
        rng.shuffle(images)

        num_classes = len(self.labelmap) if not self.is_multitask else sum(len(x) for x in self.labelmap.values())
        total_counter = collections.Counter({i: num_min_samples_per_class for i in range(num_classes)})
        sampled_images = []
        for image in images:
            counts = collections.Counter([self._get_cid(c) for c in image.labels] if not self.is_multitask else [self._get_cid(c, t) for t, t_labels in image.labels.items() for c in t_labels])
            if set((+total_counter).keys()) & set(counts.keys()):
                total_counter -= counts
                sampled_images.append(image)

            if not +total_counter:
                break

        if +total_counter:
            raise RuntimeError(f"Couldn't find {num_min_samples_per_class} samples for some classes: {+total_counter}")

        return DatasetManifest(sampled_images, self.labelmap, self.data_type)

    @staticmethod
    def merge(*args, flavor: int = 0):
        """
        merge multiple data manifests into one.

        Args:
            args: manifests to be merged
            flavor: flavor of dataset merge (not difference for captioning)
                0: merge manifests of the same type and the same labelmap (for multitask, it should be same set of tasks and same labelmap for each task)
                1: concat manifests of the same type, the new labelmap are concats of all labelmaps in all manifest (for multitask, duplicate task names are not allowed)
        """

        assert len(args) >= 1, 'less than one manifests provided, not possible to merged.'
        assert all(args), '"None" manifest found'

        if len(args) == 1:
            logger.warning('Only one manifest provided. Nothing to be merged.')
            return args[0]

        if any([isinstance(x.data_type, dict) for x in args]):
            assert all([isinstance(x.data_type, dict) for x in args]), 'Cannot merge multitask manifest and single task manifest'
        else:
            assert len(set([x.data_type for x in args])) == 1, 'All manifests must be of the same data type'

        if flavor == 0:
            return DatasetManifest._merge_with_same_labelmap(*args)
        elif flavor == 1:
            return DatasetManifest._merge_with_concat(*args)
        else:
            raise ValueError(f'Unknown flavor {flavor}.')

    @staticmethod
    def _merge_with_same_labelmap(*args):
        for i in range(len(args)):
            if i > 0 and args[i].labelmap != args[i - 1].labelmap:
                raise ValueError('labelmap must be the same for all manifests.')
            if i > 0 and args[i].data_type != args[i - 1].data_type:
                raise ValueError('Data type must be the same for all manifests.')

        images = [y for x in args for y in x.images]

        return DatasetManifest(images, args[0].labelmap, args[0].data_type)

    @staticmethod
    def _merge_with_concat(*args):
        data_type = args[0].data_type

        if data_type == DatasetTypes.IMCAP:
            return DatasetManifest._merge_with_same_labelmap(args)

        if isinstance(data_type, dict):  # multitask
            labelmap = {}
            data_types = {}
            for manifest in args:
                for k, v in manifest.labelmap.items():
                    if k in labelmap:
                        raise ValueError(f'Failed to merge dataset manifests, as due to task with name {k} exists in more than one manifest.')

                    labelmap[k] = v

                for k, v in manifest.data_type.items():
                    data_types[k] = v

            return DatasetManifest([y for x in args for y in x.images], labelmap, data_types)

        labelmap = []
        images = []

        for manifest in args:
            label_offset = len(labelmap)
            for img_manifest in manifest.images:
                new_img_manifest = copy.deepcopy(img_manifest)
                if DatasetTypes.is_classification(data_type):
                    new_img_manifest.labels = [x + label_offset for x in new_img_manifest.labels]
                elif data_type == DatasetTypes.OD:
                    for label in new_img_manifest.labels:
                        label[0] += label_offset
                else:
                    raise ValueError(f'Unsupported type in merging {data_type}')

                images.append(new_img_manifest)
            labelmap.extend(manifest.labelmap)

        return DatasetManifest(images, labelmap, data_type)

    @staticmethod
    def create_multitask_manifest(manifest_by_task: dict):
        """
        Merge several manifests into a multitask dataset in a naive way, assuming images from different manifests are independent different images.
        Args:
            manifest_by_task (dict): manifest by task name

        Returns:
            a merged multitask manifest
        """

        task_names = sorted(list(manifest_by_task.keys()))
        images = []
        for task_name in task_names:
            for img in manifest_by_task[task_name].images:
                new_img = copy.deepcopy(img)
                new_img.labels = {task_name: new_img.labels}
                images.append(new_img)

        labelmap = {task_name: manifest_by_task[task_name].labelmap for task_name in task_names}
        data_types = {task_name: manifest_by_task[task_name].data_type for task_name in task_names}

        return DatasetManifest(images, labelmap, data_types)


def _generate_multitask_dataset_manifest(manifest_by_task: Dict[str, DatasetManifest]):
    images_by_id = {}
    for task_name, task_manifest in manifest_by_task.items():
        if not task_manifest:
            continue
        for image in task_manifest.images:
            if image.id not in images_by_id:
                multi_task_image_manifest = ImageDataManifest(image.id, image.img_path, image.width, image.height, {task_name: image.labels})
                images_by_id[image.id] = multi_task_image_manifest
            else:
                images_by_id[image.id].labels[task_name] = image.labels

    if not images_by_id:
        return None

    labelmap_by_task = {k: manifest.labelmap for k, manifest in manifest_by_task.items()}
    dataset_types_by_task = {k: manifest.data_type for k, manifest in manifest_by_task.items()}
    return DatasetManifest([v for v in images_by_id.values()], labelmap_by_task, dataset_types_by_task)


class IrisManifestAdaptor:
    """
    Adaptor for generating dataset manifest from iris format
    """

    @staticmethod
    def create_dataset_manifest(dataset_info, usage: str, container_sas_or_root_dir: str = None):
        """

        Args:
            dataset_info (MultiTaskDatasetInfo or .DatasetInfo):  dataset info
            usage (str): which usage of data to construct
            container_sas_or_root_dir (str): sas url if the data is store in a azure blob container, or a local root dir
        """
        assert dataset_info
        assert usage

        if dataset_info.type == DatasetTypes.IMCAP:
            raise ValueError('Iris format is not supported for image caption task, please use COCO format!')
        if isinstance(dataset_info, MultiTaskDatasetInfo):
            dataset_manifest_by_task = {k: IrisManifestAdaptor.create_dataset_manifest(task_info, usage, container_sas_or_root_dir) for k, task_info in dataset_info.sub_task_infos.items()}
            return _generate_multitask_dataset_manifest(dataset_manifest_by_task)
        if usage not in dataset_info.index_files:
            return None

        file_reader = FileReader()

        dataset_info = copy.deepcopy(dataset_info)
        get_full_sas_or_path = _construct_full_sas_or_path_generator(container_sas_or_root_dir, dataset_info.root_folder)

        max_index = 0
        labelmap = None
        if not dataset_info.labelmap:
            logger.warning(f'{dataset_info.name}: labelmap is missing!')
        else:
            # read tag names
            with file_reader.open(get_full_sas_or_path(dataset_info.labelmap)) as file_in:
                labelmap = [purge_line(line) for line in file_in if purge_line(line) != '']

        # read image width and height
        img_wh = None
        if dataset_info.image_metadata_path:
            img_wh = IrisManifestAdaptor._load_img_width_and_height(file_reader, get_full_sas_or_path(dataset_info.image_metadata_path))

        # read image index files
        images = []
        with file_reader.open(get_full_sas_or_path(dataset_info.index_files[usage])) as file_in:
            for line in file_in:
                line = purge_line(line)
                if not line:
                    continue
                parts = line.rsplit(' ', maxsplit=1)  # assumption: only the image file path can have spaces
                img_path = parts[0]
                label_or_label_file = parts[1] if len(parts) == 2 else None

                w, h = img_wh[img_path] if img_wh else (None, None)
                if DatasetTypes.is_classification(dataset_info.type):
                    img_labels = [int(x) for x in label_or_label_file.split(',')] if label_or_label_file else []
                else:
                    img_labels = IrisManifestAdaptor._load_detection_labels_from_file(file_reader, get_full_sas_or_path(label_or_label_file)) if label_or_label_file else []

                if not labelmap and img_labels:
                    c_indices = [x[0] for x in img_labels] if isinstance(img_labels[0], list) else img_labels
                    max_index = max(max(c_indices), max_index)

                images.append(ImageDataManifest(img_path, get_full_sas_or_path(img_path), w, h, img_labels))

            if not labelmap:
                labelmap = [str(x) for x in range(max_index + 1)]
            file_reader.close()
        return DatasetManifest(images, labelmap, dataset_info.type)

    @staticmethod
    def _load_img_width_and_height(file_reader, file_path):
        img_wh = dict()
        with file_reader.open(file_path) as file_in:
            for line in file_in:
                line = purge_line(line)
                if line == '':
                    continue
                location, w, h = line.split()
                img_wh[location] = (int(w), int(h))

        return img_wh

    @staticmethod
    def _load_detection_labels_from_file(file_reader, image_label_file_path):

        with file_reader.open(image_label_file_path) as label_in:
            label_lines = [purge_line(line) for line in label_in]

        img_labels = []
        for label_line in label_lines:
            parts = label_line.split()

            assert len(parts) == 5  # regions
            box = [float(p) for p in parts]
            box[0] = int(box[0])
            img_labels.append(box)

        return img_labels


class CocoManifestAdaptor:
    """
    Adaptor for generating manifest from coco format

    image paths should be stored under 'file_name'
    """

    @staticmethod
    def create_dataset_manifest(coco_file_path_or_url, data_type, container_sas_or_root_dir: str = None):
        """ construct a dataset manifest out of coco file
        Args:
            coco_file_path_or_url (str or dict): path or url to coco file. dict if multitask
            data_type (str or dict): type of dataset. dict if multitask
            container_sas_or_root_dir (str): container sas if resources are store in blob container, or a local dir
        """
        if not coco_file_path_or_url:
            return None

        assert data_type

        if isinstance(coco_file_path_or_url, dict):
            assert isinstance(data_type, dict)
            dataset_manifest_by_task = {k: CocoManifestAdaptor.create_dataset_manifest(coco_file_path_or_url[k], data_type[k], container_sas_or_root_dir)
                                        for k in coco_file_path_or_url}

            return _generate_multitask_dataset_manifest(dataset_manifest_by_task)

        get_full_sas_or_path = _construct_full_sas_or_path_generator(container_sas_or_root_dir)

        file_reader = FileReader()
        # read image index files
        with file_reader.open(coco_file_path_or_url if is_url(coco_file_path_or_url) else get_full_sas_or_path(coco_file_path_or_url)) as file_in:
            coco_manifest = json.load(file_in)

        file_reader.close()

        if data_type == DatasetTypes.IMCAP:
            images_by_id = {img['id']: ImageDataManifest(img['id'], get_full_sas_or_path(img['file_name']), img.get('width'), img.get('height'), []) for img in coco_manifest['images']}
            for annotation in coco_manifest['annotations']:
                images_by_id[annotation['image_id']].labels.append(annotation['caption'])
            images = [x for x in images_by_id.values()]
            return DatasetManifest(images, None, data_type)

        images_by_id = {img['id']: ImageDataManifest(img['id'], get_full_sas_or_path(img['file_name']), img['width'], img['height'], []) for img in coco_manifest['images']}

        cate_id_name = [(cate['id'], cate['name']) for cate in coco_manifest['categories']]
        cate_id_name.sort(key=lambda x: x[0])
        label_id_to_pos = {x[0]: i for i, x in enumerate(cate_id_name)}
        labelmap = [x[1] for x in cate_id_name]

        bbox_format = coco_manifest.get('bbox_format', BBoxFormat.LTWH)
        BBoxFormat.validate(bbox_format)

        for annotation in coco_manifest['annotations']:
            c_id = label_id_to_pos[annotation['category_id']]
            if 'bbox' in annotation:
                bbox = annotation['bbox']
                if bbox_format == BBoxFormat.LTWH:
                    bbox = [bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]]
                label = [c_id] + bbox
            else:
                label = c_id
            images_by_id[annotation['image_id']].labels.append(label)
        images = [x for x in images_by_id.values()]
        images.sort(key=lambda x: x.id)

        return DatasetManifest(images, labelmap, data_type)



from .common import DatasetRegistry, Usages, DatasetTypes, CocoManifestAdaptor, IrisManifestAdaptor, DatasetManifest, DatasetInfo, ManifestDataset, BaseDatasetInfo, MultiTaskDatasetInfo
from .resources import DatasetHub

__all__ = ['DatasetRegistry', 'Usages', 'DatasetTypes', 'CocoManifestAdaptor', 'IrisManifestAdaptor', 'DatasetManifest', 'DatasetInfo', 'ManifestDataset', 'BaseDatasetInfo', 'MultiTaskDatasetInfo',
           'DatasetHub']



import argparse
import logging
import os
from PIL import Image

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser('Generate image metainfo needed for object detection datasets.')
    parser.add_argument('index_files', nargs='+', help='index files.')
    parser.add_argument('--output', type=str, help='meta file name', required=True)
    parser.add_argument('--base_folder', type=str, help='base_folder', default='.\\')

    args = parser.parse_args()
    with open(os.path.join(args.base_folder, args.output), 'w') as out_file:
        for file in args.index_files:
            with open(os.path.join(args.base_folder, file), 'r') as file_in:
                for line in file_in:
                    img_id, _ = line.strip().split(' ')
                    img_loc = os.path.join(args.base_folder, img_id.replace('.zip@', '\\'))
                    img = Image.open(img_loc)
                    w, h = img.size
                    out_file.write(f'{img_id} {w} {h}\n')


if __name__ == '__main__':
    main()



from .constants import Usages, DatasetTypes, Formats


class DatasetInfoFactory:
    @staticmethod
    def create(data_info_dict):
        if DatasetInfoFactory.is_multitask(data_info_dict['type']):
            return MultiTaskDatasetInfo(data_info_dict)
        return DatasetInfo(data_info_dict)

    @staticmethod
    def is_multitask(task_type):
        return 'multitask' in task_type


class BaseDatasetInfo:
    """
    Info fields common to both all datasets regardless of whether it is coco or iris, single task or multitask
    """

    def __init__(self, dataset_info_dict):
        self.name = dataset_info_dict['name']
        self.version = dataset_info_dict['version']
        self.type = dataset_info_dict['type']
        self.root_folder = dataset_info_dict['root_folder']
        self.description = dataset_info_dict.get('description', '')
        self.data_format = dataset_info_dict.get('format', Formats.IRIS)


class DatasetInfo(BaseDatasetInfo):

    def __init__(self, dataset_info_dict):
        assert dataset_info_dict['type'] in DatasetTypes.VALID_TYPES
        assert not DatasetInfoFactory.is_multitask(dataset_info_dict['type'])
        super(DatasetInfo, self).__init__(dataset_info_dict)

        self.index_files = dict()
        self.files_for_local_usage = dict()
        for usage in [Usages.TRAIN_PURPOSE, Usages.VAL_PURPOSE, Usages.TEST_PURPOSE]:
            if usage in dataset_info_dict:
                self.index_files[usage] = dataset_info_dict[usage]['index_path']
                self.files_for_local_usage[usage] = dataset_info_dict[usage]['files_for_local_usage']

        self.box_out_of_image = dataset_info_dict.get('box_out_of_image', False)

        # Below are needed for iris format only. As both image h and w and labelmaps are included in the coco annotation files
        self.labelmap = dataset_info_dict.get('labelmap')
        self.image_metadata_path = dataset_info_dict.get('image_metadata_path')

    @property
    def train_path(self):
        return self.index_files[Usages.TRAIN_PURPOSE] if Usages.TRAIN_PURPOSE in self.index_files else None

    @property
    def val_path(self):
        return self.index_files[Usages.VAL_PURPOSE] if Usages.VAL_PURPOSE in self.index_files else None

    @property
    def test_path(self):
        return self.index_files[Usages.TEST_PURPOSE] if Usages.TEST_PURPOSE in self.index_files else None

    @property
    def train_support_files(self):
        """Path to the files which are referenced by the train dataset file"""

        return self.files_for_local_usage[Usages.TRAIN_PURPOSE] if Usages.TRAIN_PURPOSE in self.index_files else []

    @property
    def val_support_files(self):
        """Path to the files which are referenced by the validation dataset file"""

        return self.files_for_local_usage[Usages.VAL_PURPOSE] if Usages.VAL_PURPOSE in self.index_files else []

    @property
    def test_support_files(self):
        """Path to the files which are referenced by the test dataset file"""

        return self.files_for_local_usage[Usages.TEST_PURPOSE] if Usages.TEST_PURPOSE in self.index_files else []


class MultiTaskDatasetInfo(BaseDatasetInfo):
    def __init__(self, dataset_info_dict):
        assert 'tasks' in dataset_info_dict
        assert DatasetInfoFactory.is_multitask(dataset_info_dict['type'])

        super(MultiTaskDatasetInfo, self).__init__(dataset_info_dict)

        tasks = dataset_info_dict['tasks']
        info_dict = {}
        for task_name, task_info in tasks.items():
            info_dict[task_name] = DatasetInfo({**dataset_info_dict, **task_info})

        self.sub_task_infos = info_dict

    @property
    def task_names(self):
        return list(self.sub_task_infos.keys())

    def get_task_dataset_info(self, task_name: str):
        return self.sub_task_infos[task_name]

    @property
    def train_support_files(self):
        """Path to the files which are referenced by the train dataset file"""
        return list(set([x for task_info in self.sub_task_infos.values() for x in task_info.train_support_files]))

    @property
    def val_support_files(self):
        """Path to the files which are referenced by the validation dataset file"""

        return list(set([x for task_info in self.sub_task_infos.values() for x in task_info.val_support_files]))

    @property
    def test_support_files(self):
        """Path to the files which are referenced by the validation dataset file"""

        return list(set([x for task_info in self.sub_task_infos.values() for x in task_info.test_support_files]))



import os
import zipfile
from urllib import parse as urlparse
from urllib.parse import quote
from urllib.request import urlopen


def is_url(candidate: str):
    """
    necessary condition to be a url (not sufficient)
    Args:
        candidate (str):

    Returns:
        whether it could be a sas url or not

    """
    try:
        result = urlparse.urlparse(candidate)
        return result.scheme and result.netloc
    except ValueError:
        return False


class MultiProcessZipFile:
    """ZipFile which is readable from multi processes"""

    def __init__(self, filename):
        self.filename = filename
        self.zipfiles = {}

    def open(self, file):
        if os.getpid() not in self.zipfiles:
            self.zipfiles[os.getpid()] = zipfile.ZipFile(self.filename)
        return self.zipfiles[os.getpid()].open(file)

    def close(self):
        for z in self.zipfiles.values():
            z.close()
        self.zipfiles = {}

    def __getstate__(self):
        return {'filename': self.filename}

    def __setstate__(self, state):
        self.filename = state['filename']
        self.zipfiles = {}


class FileReader:
    """Reader to support <zip_filename>@<entry_name> style filename."""

    def __init__(self):
        self.zip_files = {}

    def open(self, name, mode='r'):
        # read file from url
        if is_url(name):
            return urlopen(self._encode_non_ascii(name))

        # read file from local zip: <zip_filename>@<entry_name>, e.g. images.zip@1.jpg
        if '@' in name:
            zip_path, file_path = name.split('@')
            if zip_path not in self.zip_files:
                self.zip_files[zip_path] = MultiProcessZipFile(zip_path)
            return self.zip_files[zip_path].open(file_path)

        # read file from local dir
        return open(name, mode)

    def close(self):
        for zip_file in self.zip_files.values():
            zip_file.close()
        self.zip_files = {}

    @staticmethod
    def _encode_non_ascii(s):
        return ''.join([c if ord(c) < 128 else quote(c) for c in s])



from abc import ABC, abstractmethod

from .dataset_info import BaseDatasetInfo


class BaseDataset(ABC):
    def __init__(self, dataset_info: BaseDatasetInfo):

        self.dataset_info = dataset_info

    def __getitem__(self, idx):
        """ iterate through the dataset

        Args:
            idx: can be a single index or range

        Returns:
            requested sample(s)
        """

        if isinstance(idx, int):
            if idx >= self.__len__():
                raise IndexError
            return self._get_single_item(idx)

        stop = min(self.__len__(), idx.stop)
        return [self.__getitem__(i) for i in range(idx.start, stop, idx.step)] if idx.step else [self.__getitem__(i) for i in range(idx.start, stop)]

    @property
    @abstractmethod
    def labels(self):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def _get_single_item(self, index):
        pass

    @abstractmethod
    def close(self):
        """ release resources
        """
        pass



import logging
import os
import pathlib
from typing import List

import requests
import shutil
import tempfile
import tenacity
from urllib import parse as urlparse

from .dataset_registry import DatasetRegistry
from .dataset_info import DatasetInfo, DatasetInfoFactory
from .constants import Usages
from .util import is_url

logger = logging.getLogger(__name__)


class DownloadedDatasetsResources:
    """Wrapper class to make sure the temporary directory is removed."""

    def __init__(self, base_dirs: List[pathlib.Path]):
        self.base_dirs = base_dirs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        for base_dir in self.base_dirs:
            if os.path.isdir(base_dir):
                logger.info(f'Removing folder: {base_dir}.')
                shutil.rmtree(base_dir)

    @staticmethod
    def merge(r1, r2):
        assert r1
        assert r2

        return DownloadedDatasetsResources(r1.base_dirs + r2.base_dirs)


class DatasetDownloader:
    def __init__(self, dataset_sas_url: str, dataset_registry: DatasetRegistry):
        assert dataset_sas_url
        assert dataset_registry

        if not is_url(dataset_sas_url):
            raise RuntimeError('An url to the dataset should be provided.')

        self._dataset_sas_url = dataset_sas_url
        self._registry = dataset_registry

    def download(self, name: str, version: int = None, target_dir: str = None, purposes=[Usages.TRAIN_PURPOSE, Usages.VAL_PURPOSE, Usages.TEST_PURPOSE]):
        assert purposes

        dataset_info = self._registry.get_dataset_info(name, version)
        if not dataset_info:
            raise RuntimeError(f'No dataset matched for the specified condition: {name} ({version})')

        target_dir = pathlib.Path(tempfile.mkdtemp()) if target_dir is None else pathlib.Path(target_dir)
        target_dir = target_dir / pathlib.Path(dataset_info.root_folder)
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)

        if DatasetInfoFactory.is_multitask(dataset_info.type):
            for subtask_info in dataset_info.sub_task_infos.values():
                self._download(subtask_info, target_dir, purposes)
        else:
            self._download(dataset_info, target_dir, purposes)

        return DownloadedDatasetsResources([target_dir])

    def _download(self, dataset_info: DatasetInfo, target_dir, purposes):
        files_to_download = set()

        for usage in purposes:
            if usage in dataset_info.index_files:
                files_to_download.add(os.path.join(dataset_info.root_folder, dataset_info.index_files[usage]))
            if usage in dataset_info.files_for_local_usage:
                files_to_download.update([os.path.join(dataset_info.root_folder, x) for x in dataset_info.files_for_local_usage[usage]])

        if dataset_info.labelmap:
            files_to_download.add(os.path.join(dataset_info.root_folder, dataset_info.labelmap))

        if dataset_info.image_metadata_path:
            files_to_download.add(os.path.join(dataset_info.root_folder, dataset_info.image_metadata_path))

        self._download_files(files_to_download, target_dir)

    def _download_files(self, file_paths, target_dir: pathlib.Path):
        parts = urlparse.urlparse(self._dataset_sas_url)
        for file_path in file_paths:
            path = os.path.join(parts[2], file_path).replace('\\', '/')
            url = urlparse.urlunparse((parts[0], parts[1], path, parts[3], parts[4], parts[5]))
            target_file_path = target_dir / pathlib.Path(file_path).name
            if os.path.exists(target_file_path):
                logger.info(f'{target_file_path} exists. Skip downloading.')
                continue

            self._download_file(url, target_file_path)

    @tenacity.retry(stop=tenacity.stop_after_attempt(3))
    def _download_file(self, url, filepath):
        logger.info(f'Downloading from {url} to {filepath.absolute()}.')
        with requests.get(url, stream=True, allow_redirects=True, timeout=60) as r:
            if r.status_code > 200:
                raise RuntimeError(f'Failed in downloading from {url}, status code {r.status_code}.')

            with open(filepath, 'wb') as f:
                shutil.copyfileobj(r.raw, f, length=4194304)



import base64
import io
import logging
import os

from vision_datasets.common.image_loader import PILImageLoader

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

TSV_FORMAT_LTRB = 'ltrb'
TSV_FORMAT_LTWH_NORM = 'ltwh-normalized'


def zip_folder(folder_name):
    import zipfile

    logger.info(f'zipping {folder_name}.')

    zip_file = zipfile.ZipFile(f'{folder_name}.zip', 'w', zipfile.ZIP_STORED)
    i = 0
    for root, dirs, files in os.walk(folder_name):
        for file in files:
            if i % 5000 == 0:
                logger.info(f'zipped {i} images')

            zip_file.write(os.path.join(root, file))
            i += 1

    zip_file.close()


def decode64_to_pil(img_b64_str):
    assert img_b64_str

    return PILImageLoader.load_from_stream(io.BytesIO(base64.b64decode(img_b64_str)))


def guess_encoding(csv_file):
    """guess the encoding of the given file https://stackoverflow.com/a/33981557/
    """
    import io
    import locale

    with io.open(csv_file, 'rb') as f:
        data = f.read(5)
    if data.startswith(b'\xEF\xBB\xBF'):  # UTF-8 with a "BOM"
        return 'utf-8-sig'
    elif data.startswith(b'\xFF\xFE') or data.startswith(b"\xFE\xFF"):
        return 'utf-16'
    else:  # in Windows, guessing utf-8 doesn't work, so we have to try
        # noinspection PyBroadException
        try:
            with io.open(csv_file, encoding='utf-8') as f:
                f.read(222222)
                return 'utf-8'
        except Exception:
            return locale.getdefaultlocale()[1]


def verify_and_correct_box_or_none(lp, box, data_format, img_w, img_h):
    error_msg = f'{lp} Illegal box [{", ".join([str(x) for x in box])}], img wxh: {img_w}, {img_h}'
    if len([x for x in box if x < 0]) > 0:
        logger.error(f'{error_msg}. Skip this box.')
        return None

    if data_format == TSV_FORMAT_LTWH_NORM:
        box[2] = int((box[0] + box[2]) * img_w)
        box[3] = int((box[1] + box[3]) * img_h)
        box[0] = int(box[0] * img_w)
        box[1] = int(box[1] * img_h)

    boundary_ratio_limit = 1.02
    if box[0] >= img_w or box[1] >= img_h or box[2] / img_w > boundary_ratio_limit \
            or box[3] / img_h > boundary_ratio_limit or box[0] >= box[2] or box[1] >= box[3]:
        logger.error(f'{error_msg}. Skip this box.')
        return None

    box[2] = min(box[2], img_w)
    box[3] = min(box[3], img_h)

    return box



from PIL import ImageFile
from abc import ABC
import torch
from inspect import signature


def _identity(*args):
    return args


class _ImageOnlyTransform:
    def __init__(self, transform):
        self._transform = transform

    def __call__(self, image, boxes):
        return self._transform(image), boxes


class Dataset(torch.utils.data.Dataset, ABC):
    def __init__(self, transform=None):
        super().__init__()
        self.transform = transform
        # Work around for corrupted files in datasets
        ImageFile.LOAD_TRUNCATED_IMAGES = True

    @property
    def labels(self):
        """Returns a list of labels."""
        raise NotImplementedError

    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, val):
        if not val:
            self._transform = _identity
        elif len(signature(val).parameters) == 1:
            self._transform = _ImageOnlyTransform(val)
        else:
            self._transform = val

    def close(self):
        """Release the resources allocated for this dataset."""
        pass

    def __getitem__(self, index):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError



import logging

from .dataset import Dataset

logger = logging.getLogger(__name__)


class TorchDataset(Dataset):
    """
    Dataset class used for pytorch training
    """

    def __init__(self, manifest_dataset, transform=None):
        Dataset.__init__(self, transform)
        self.dataset = manifest_dataset

    @property
    def labels(self):
        return self.dataset.labels

    @property
    def dataset_resources(self):
        return self.dataset.dataset_resources

    @property
    def dataset_info(self):
        return self.dataset.dataset_info

    def __getitem__(self, index):
        if isinstance(index, int):
            image, target, idx_str = self.dataset[index]
            image, target = self.transform(image, target)
            return image, target, idx_str
        else:
            return [self.transform(img, target) + (idx,) for img, target, idx in self.dataset[index]]

    def __len__(self):
        return len(self.dataset)

    def close(self):
        self.dataset.close()



import argparse
import datetime
import json
import pathlib
import os
import zipfile

from vision_datasets import DatasetTypes

TRAIN_USAGE = 'train'
VAL_USAGE = 'val'
TEST_USAGE = 'test'


def zipdir(path, ziph):
    # ziph is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in files:
            k = os.path.relpath(os.path.join(root, file), path)
            ziph.write(os.path.join(root, file), k)


def create_argparse():
    parser = argparse.ArgumentParser('Prepare the annotation files, data reg json, and zip files for vision-datasets following iris format.')
    parser.add_argument('name', type=str, help="Dataset name.")
    parser.add_argument('--type', '-t', type=str, default=DatasetTypes.IC_MULTICLASS, help="type of dataset.", choices=[DatasetTypes.IC_MULTICLASS, DatasetTypes.IC_MULTILABEL])
    parser.add_argument('--description', '-d', type=str, help="Dataset description.", required=True)
    parser.add_argument('--contact', '-c', type=str, help="contact person.", required=False)
    parser.add_argument('--train_folder', '-tr', type=pathlib.Path, help="Folder including training images.")
    parser.add_argument('--val_folder', '-v', type=pathlib.Path, help="Folder including validation images.")
    parser.add_argument('--test_folder', '-te', type=pathlib.Path, help="Folder including test images.")
    return parser


def main():
    parser = create_argparse()
    args = parser.parse_args()

    labelmap_file = 'labels.txt'
    today = datetime.datetime.now().strftime('%Y%m%d')
    reg_json = {
        'name': args.name,
        'description': args.description,
        'contact': args.contact,
        'version': 1,
        "type": args.type,
        "root_folder": f"classification/{args.name.replace('-', '_')}_{today}",
        "labelmap": labelmap_file,
    }

    folder_by_usage = {
        TRAIN_USAGE: args.train_folder,
        VAL_USAGE: args.val_folder,
        TEST_USAGE: args.test_folder
    }

    classes = os.listdir(folder_by_usage[TRAIN_USAGE])
    reg_json['num_classes'] = len(classes)
    with open(labelmap_file, 'w') as label_out:
        label_out.write('\n'.join(classes))

    n_images = {usage: 0 for usage in folder_by_usage.keys()}

    for usage, folder in folder_by_usage.items():
        if not folder:
            continue

        with open(f'{usage}.txt', 'w') as index_file:
            for i, c in enumerate(classes):
                for img_file in (folder / c).iterdir():
                    img_path = img_file.as_posix()
                    img_path = str(img_path).replace(f'{folder}/', f'{folder}.zip@')
                    index_file.write(f'{img_path} {i}\n')
                    n_images[usage] += 1

        with zipfile.ZipFile(f'{folder}.zip', 'w') as zipf:
            zipdir(folder, zipf)

        reg_json[usage] = {
            "index_path": f"{usage}.txt",
            "files_for_local_usage": [
                f"{folder}.zip"
            ],
            "num_images": n_images[usage]
        }

    pathlib.Path('reg.json').write_text(json.dumps(reg_json, indent=2))


if __name__ == '__main__':
    main()



"""
Converts YOLO Darknet TXT format to iris format
"""

import argparse
import os
import pathlib
import zipfile

from datetime import datetime
from PIL import Image
from tqdm import tqdm
from typing import List

IMAGE_FILE_EXTENSION = '.jpg'


def create_arg_parser():
    parser = argparse.ArgumentParser(description='Convert YOLO Darknet TXT format to iris format for an object detection dataset.')
    parser.add_argument('-i', '--input', type=pathlib.Path, required=True, help='Path to the directory of image and label data in YOLO format.')
    parser.add_argument('-o', '--output', type=pathlib.Path, help='Path to the output directory to write data to in iris format.')

    zip_parser = parser.add_mutually_exclusive_group(required=False)
    zip_parser.add_argument('--zip', dest='zip', action='store_true', help='Flag to choose to zip the image and label folders (default).')
    zip_parser.add_argument('--no-zip', dest='zip', action='store_false', help='Flag to choose NOT to zip the image and label folders.')
    parser.set_defaults(zip=True)
    return parser


def read_annotation_file(path: pathlib.Path):
    return [list(map(float, bbox.split())) for bbox in path.read_text().splitlines()]


def write_iris_annotation_file(path: pathlib.Path, bboxes: List[List]):
    path.write_text('\n'.join([' '.join(map(str, bbox)) for bbox in bboxes]))


def validate_and_correct_iris_bbox(bbox: List, image_width: int, image_height: int):
    if 0 <= bbox[1] <= image_width and 0 <= bbox[2] <= image_height:
        bbox[3], bbox[4] = min(bbox[3], image_width), min(bbox[4], image_height)
        return list(map(int, bbox))


def convert_yolo_bboxes_to_iris(yolo_bboxes: List[List[float]], image_width: int, image_height: int):
    iris_bboxes = []
    for _class, x_center, y_center, bbox_width, bbox_height in yolo_bboxes:
        iris_bbox = [_class,
                     (x_center - bbox_width / 2) * image_width,
                     (y_center - bbox_height / 2) * image_height,
                     (x_center + bbox_width / 2) * image_width,
                     (y_center + bbox_height / 2) * image_height]
        iris_bboxes += [validate_and_correct_iris_bbox(iris_bbox, image_width, image_height)]
    return list(filter(None, iris_bboxes))


def zip_directory(dirname):
    zip_file = zipfile.ZipFile(f'{dirname}.zip', 'w', zipfile.ZIP_STORED)
    os.chdir(dirname)
    for _, _, files in os.walk(dirname):
        for file in files:
            zip_file.write(file)
    zip_file.close()


def main():
    args = create_arg_parser().parse_args()
    if not args.input.is_dir():
        raise argparse.ArgumentTypeError(f'{args.input} is not a valid directory path')

    if not args.output:
        args.output = args.input.parent / f'{args.input.name}_{datetime.today().strftime("%Y%m%d")}'
    args.output.mkdir(exist_ok=True)  # create output directory if not exists

    im_meta_info_fh = (args.output / 'image_meta_info.txt').open('w')
    datasplits = [subdir.name for subdir in args.input.iterdir() if subdir.is_dir()]

    for datasplit in datasplits:
        input_image_dir, input_label_dir = args.input / datasplit / 'images', args.input / datasplit / 'labels'
        assert len(list(input_image_dir.iterdir())) == len(list(input_label_dir.iterdir()))

        output_image_dir, output_label_dir = args.output / (datasplit + '_images'), args.output / (datasplit + '_labels')
        output_image_dir.mkdir(exist_ok=True)
        output_label_dir.mkdir(exist_ok=True)
        index_fh = (args.output / (datasplit + '_images.txt')).open('w')

        for label_file in tqdm(input_label_dir.iterdir(), desc=datasplit):
            image_fname = label_file.stem + IMAGE_FILE_EXTENSION
            input_image_fp = input_image_dir / image_fname
            image_width, image_height = Image.open(input_image_fp).size
            input_image_fp.replace(output_image_dir / image_fname)  # move image to output image directory

            # write to image meta info and index file
            im_meta_info_fh.write(f'{output_image_dir.name}.zip@{image_fname} {image_width} {image_height}\n')
            index_fh.write(f'{output_image_dir.name}.zip@{image_fname} {output_label_dir.name}.zip@{label_file.name}\n')

            # convert annotation from yolo to iris format
            yolo_bboxes = read_annotation_file(label_file)
            iris_bboxes = convert_yolo_bboxes_to_iris(yolo_bboxes, image_width, image_height)
            write_iris_annotation_file(output_label_dir / label_file.name, iris_bboxes)

        if args.zip:
            zip_directory(output_image_dir)
            zip_directory(output_label_dir)

        index_fh.close()
    im_meta_info_fh.close()


if __name__ == '__main__':
    main()



import logging

from PIL import Image

logger = logging.getLogger(__name__)

# see https://exiv2.org/tags.html
ORIENTATION_EXIF_TAG = 0x0112


class PILImageLoader:
    """Load PIL image and fix image orientation using EXIF"""

    @staticmethod
    def load_from_stream(f):
        image = Image.open(f)
        img_format = image.format

        exif = image.getexif()
        orientation = exif.get(ORIENTATION_EXIF_TAG) if exif else None
        if orientation:
            # orientation is 1 based, shift to zero based and flip/transpose based on 0-based values
            orientation -= 1
            if orientation >= 4:
                image = image.transpose(Image.TRANSPOSE)
            if orientation == 2 or orientation == 3 or orientation == 6 or orientation == 7:
                image = image.transpose(Image.FLIP_TOP_BOTTOM)
            if orientation == 1 or orientation == 2 or orientation == 5 or orientation == 6:
                image = image.transpose(Image.FLIP_LEFT_RIGHT)
        image = image.convert('RGB')
        image.format = img_format
        return image

    @staticmethod
    def load_from_file(filepath):
        try:
            with open(filepath, 'rb') as f:
                return PILImageLoader.load_from_stream(f)
        except Exception:
            logger.exception(f'Failed to load an image: {filepath}')
            raise



import logging
from typing import Union, List

from ..common import DatasetRegistry, Usages
from ..common.data_manifest import DatasetManifest
from ..common.dataset_info import MultiTaskDatasetInfo

logger = logging.getLogger(__name__)


class DatasetHub(object):
    """
    Hub class for managing vision dataset resources, with a few common utilities for creating a dataset.
    This hub class works with both resources on local disk or on azure blob.
    """

    def __init__(self, dataset_json_str: Union[str, list]):
        """

        Args:
            dataset_json_str (str, list): dataset registry json, containing multiple dataset_info for different datasets, or a list of dataset reg json
            retrievable by their names, versions and usages.
        """
        assert dataset_json_str

        self.dataset_registry = DatasetRegistry(dataset_json_str)

    def create_manifest_dataset(self, container_sas: str, local_dir: str, name: str, version: int = None, usage: str = Usages.TRAIN_PURPOSE, coordinates: str = 'relative',
                                few_shot_samples_per_class=None, rnd_seed=0):
        """Create manifest dataset.
            If local_dir is provided, manifest_dataset consumes data from local disk. If data not present on local disk, it will be automatically downloaded.
            if container_sas is provided but local_dir not provided, manifest_dataset consumes data directly from container_sas.

            Note that for data stored in zipped files, they can be consumed locally without unzip. However, in blob they must be stored in unzipped folders. In this case image/label file paths can
            stay with paths to data in zipped files, as dataset class will automatically look in the folder names same with the zip file names.

        Args:
            container_sas: sas url to the container where datasets can be found/downloaded from
            local_dir: local directory where datasets can be found/downloaded to
            name: dataset name
            version: dataset version, if not specified, latest version will be returned
            usage: usage(s) of the dataset, 'train', 'val' or 'test' or a list of usages
            coordinates: format of the bounding boxes, can be 'relative' or 'absolute'
            few_shot_samples_per_class (int): get a sampled dataset with N images at most for each class (for detection and multilabel case, not guaranteed.)
            rnd_seed (int): random seed for few shot sampling

        Returns:
            an instance of dataset for local usage
        """
        result = self.create_dataset_manifest(container_sas, local_dir, name, version, usage, few_shot_samples_per_class, rnd_seed)
        if result:
            manifest, dataset_info, downloader_resources = result
        else:
            return None

        from vision_datasets import ManifestDataset
        return ManifestDataset(dataset_info, manifest, coordinates, downloader_resources)

    def create_dataset_manifest(self, container_sas: str, local_dir: str, name: str, version: int = None, usage: Union[str, List] = Usages.TRAIN_PURPOSE, few_shot_samples_per_class=None, rnd_seed=0):
        """Create dataset manifest.
            If local_dir is provided, manifest_dataset consumes data from local disk. If data not present on local disk, it will be automatically downloaded.
            if container_sas is provided but local_dir not provided, manifest_dataset consumes data directly from container_sas.

            Note that for data stored in zipped files, they can be consumed locally without unzip. However, in blob they must be stored in unzipped folders. In this case image/label file paths can
            stay with paths to data in zipped files, as dataset class will automatically look in the folder names same with the zip file names.

        Args:
            container_sas: sas url to the container where datasets can be found/downloaded from
            local_dir: local directory where datasets can be found/downloaded to
            name: dataset name
            version: dataset version, if not specified, latest version will be returned
            usage: usage(s) of the dataset, 'train', 'val' or 'test' or a list of usages
            few_shot_samples_per_class (int): get a sampled dataset with N images at most for each class (for detection and multilabel case, not guaranteed.)
            rnd_seed (int): random seed for few shot sampling

        Returns:
            dataset manifest, dataset_info, downloaded_resources, if dataset exists, else None
        """
        assert container_sas or local_dir
        assert name
        assert usage

        from ..common.dataset_downloader import DatasetDownloader

        usages = usage if isinstance(usage, list) else [usage]

        manifest = None
        downloader_resources = None
        dataset_info = self.dataset_registry.get_dataset_info(name, version)
        if dataset_info is None:
            logger.warning(f'Dataset with {name} and version {version} not found.')
            return None

        if isinstance(dataset_info, MultiTaskDatasetInfo):
            for task_info in dataset_info.sub_task_infos.values():
                task_info.index_files = {usage: task_info.index_files[usage] for usage in usages if usage in task_info.index_files}
        else:
            dataset_info.index_files = {usage: dataset_info.index_files[usage] for usage in usages if usage in dataset_info.index_files}

        if container_sas and local_dir:
            downloader = DatasetDownloader(container_sas, self.dataset_registry)
            downloader_resources_usage = downloader.download(name, version, local_dir, usages)
        else:
            downloader_resources_usage = None

        for usage in usages:
            manifest_usage = DatasetManifest.create_dataset_manifest(dataset_info, usage, local_dir or container_sas)
            if manifest_usage:
                manifest = DatasetManifest.merge(manifest, manifest_usage) if manifest else manifest_usage

            if downloader_resources_usage:
                from ..common.dataset_downloader import DownloadedDatasetsResources
                downloader_resources = DownloadedDatasetsResources.merge(downloader_resources, downloader_resources_usage) if downloader_resources else downloader_resources_usage

        if manifest is None:
            return None

        if few_shot_samples_per_class:
            original_img_cnt = len(manifest.images)
            manifest = manifest.sample_few_shot_subset(few_shot_samples_per_class, rnd_seed)
            logger.info(f'Create a few-shot dataset with n samples per class = {few_shot_samples_per_class}. # images: {original_img_cnt} => {len(manifest.images)}')

        return manifest, dataset_info, downloader_resources

    def list_data_version_and_types(self):
        """List all dataset names, versions and types
        """

        return self.dataset_registry.list_data_version_and_types()



"""
Converts tsv format to iris format
"""

import json
import logging
import os

from .utils import guess_encoding, decode64_to_pil, zip_folder, TSV_FORMAT_LTRB, TSV_FORMAT_LTWH_NORM, verify_and_correct_box_or_none

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def create_arg_parser():
    import argparse

    parser = argparse.ArgumentParser(description='Convert detection dataset in tsv to iris format.')
    parser.add_argument('-t', '--tsvs', required=True, nargs='+', help='Tsv files to convert.')
    parser.add_argument('-l', '--labelmap', type=str, default='labelmap.txt')
    parser.add_argument('-f', '--format', type=str, default=TSV_FORMAT_LTRB, choices=[TSV_FORMAT_LTRB, TSV_FORMAT_LTWH_NORM])
    parser.add_argument('-d', '--difficulty', type=bool, default=False, help='Include difficulty boxes or not.')
    parser.add_argument('-z', '--zip', type=bool, default=True, help='Zip the image and label folders or not.')

    return parser


def main():
    args = create_arg_parser().parse_args()
    logger.info(args.__dict__)
    image_folder_name = 'images'
    if not os.path.exists(image_folder_name):
        os.mkdir(image_folder_name)
    label_folder_name = 'labels'
    if not os.path.exists(label_folder_name):
        os.mkdir(label_folder_name)

    labelmap_exists = os.path.exists(args.labelmap)
    if labelmap_exists:
        logger.info(f'Labelmap {args.labelmap} exists.')
        with open(args.labelmap, 'r') as labelmap_in:
            label_name_to_idx = {x.strip(): i for i, x in enumerate(labelmap_in.readlines())}
            if len(label_name_to_idx) == 0:
                raise Exception(f'Empty labelmap {args.labelmap}')
    else:
        logger.info(f'Labelmap {args.labelmap} does not exist, created on the fly.')
        label_name_to_idx = dict()

    img_meta_data_file_name = 'image_meta_data.txt'
    with open(img_meta_data_file_name, 'w') as image_meta_data_out:
        for tsv_file_name in args.tsvs:
            index_file_name = os.path.splitext(os.path.basename(tsv_file_name))[0] + '.txt'
            line_idx = 1
            with open(index_file_name, 'w') as index_file_out:
                with open(tsv_file_name, 'r', encoding=guess_encoding(tsv_file_name)) as file_in:
                    logger.info(f'Processing {tsv_file_name}.')
                    for img_info in file_in:
                        img_id, labels_json, img_b64 = img_info.split('\t')

                        img = decode64_to_pil(img_b64)
                        w, h = img.size

                        # image data => image file
                        img_file_name = img_id + '.' + img.format
                        img.save(os.path.join(image_folder_name, img_file_name), img.format)

                        image_path_id = f'{image_folder_name}.zip@{img_file_name}'

                        # image size => meta data file
                        image_meta_data_out.write(f'{image_path_id} {w} {h}\n')

                        # image info => index file
                        img_label_file_name = img_id + '.txt'
                        index_file_out.write(f'{image_path_id} {label_folder_name}.zip@{img_label_file_name}\n')

                        lp = f'File: {tsv_file_name}, Line {line_idx}: '
                        # labels => files
                        with open(os.path.join(label_folder_name, img_label_file_name), 'w') as label_out:
                            labels = json.loads(labels_json)
                            for label in labels:
                                difficulty = label['diff'] if 'diff' in label else 0
                                if difficulty > 0 and not args.difficulty:
                                    continue

                                if labelmap_exists and label['class'] not in label_name_to_idx:
                                    raise Exception(f'{lp}Illegal class {label["class"]}, not in provided labelmap.')

                                if label['class'] not in label_name_to_idx:
                                    label_name_to_idx[label['class']] = len(label_name_to_idx)

                                label_idx = label_name_to_idx[label['class']]
                                box = verify_and_correct_box_or_none(lp, label['rect'], args.format, w, h)
                                if box is None:
                                    continue

                                label_out.write(f'{label_idx} {box[0]} {box[1]} {box[2]} {box[3]}\n')

                        line_idx += 1

    if not labelmap_exists:
        logger.info(f'Write labelmap to {args.labelmap}')
        with open(args.labelmap, 'w') as labelmap_out:
            idx_to_labels = {label_name_to_idx[key]: key for key in label_name_to_idx}
            for i in range(len(idx_to_labels)):
                labelmap_out.write(idx_to_labels[i] + '\n')

    if args.zip:
        logger.info(f'Zip folder "{image_folder_name}".')
        zip_folder(image_folder_name)
        logger.info(f'Zip folder "{label_folder_name}".')
        zip_folder(label_folder_name)


if __name__ == '__main__':
    main()



import json
import logging
import os
import pathlib

from vision_datasets import DatasetHub, Usages
from vision_datasets.common.manifest_dataset import DetectionAsClassificationDataset

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def create_arg_parser():
    import argparse

    parser = argparse.ArgumentParser(description='Convert OD dataset to ic dataset.')
    parser.add_argument('-n', '--name', type=str, required=True, help='dataset name')
    parser.add_argument('-r', '--reg_json_path', type=str, default=None, help="dataset registration json path.", required=True)
    parser.add_argument('-k', '--sas', type=str, help="sas url.", required=False, default=None)
    parser.add_argument('-l', '--local_folder', type=str, help="detection dataset folder.", required=False)
    parser.add_argument('-o', '--output_folder', type=str, required=True, help='target folder of the converted classification dataset')
    parser.add_argument('-zb', '--zoom_ratio_bounds', type=str, required=False,
                        help='lower and bound of the ratio that box height and width can expand (>1) or shrink (0-1), during cropping, e.g, 0.8/1.2')
    parser.add_argument('-sb', '--shift_relative_bounds', type=str, required=False,
                        help='lower/upper bounds of relative ratio wrt box width and height that a box can shift, during cropping, e.g., "-0.3/0.1"')
    parser.add_argument('-s', '--rnd_seed', type=int, required=False, help='random see for box expansion/shrink/shifting.', default=0)

    return parser


def main():
    arg_parser = create_arg_parser()
    args = arg_parser.parse_args()
    dataset_resources = DatasetHub(pathlib.Path(args.reg_json_path).read_text())
    aug_params = {}
    if args.zoom_ratio_bounds:
        low, up = args.zoom_ratio_bounds.split('/')
        aug_params['zoom_ratio_bounds'] = (float(low), float(up))

    if args.shift_relative_bounds:
        low, up = args.shift_relative_bounds.split('/')
        aug_params['shift_relative_bounds'] = (float(low), float(up))

    if aug_params:
        aug_params['rnd_seed'] = args.rnd_seed

    if not os.path.exists(args.output_folder):
        os.makedirs(args.output_folder)

    if args.local_folder and not os.path.exists(args.local_folder):
        os.makedirs(args.local_folder)

    categories = None
    for phase in [Usages.TRAIN_PURPOSE, Usages.VAL_PURPOSE, Usages.TEST_PURPOSE]:
        images = []
        annotations = []

        logger.info(f'download dataset manifest for {args.name}...')
        dataset = dataset_resources.create_manifest_dataset(args.sas, args.local_folder, args.name, usage=phase, coordinates='absolute')
        if not dataset:
            logger.info(f'Skipping phase {phase}.')
            continue

        img_folder = os.path.join(args.output_folder, phase)
        if not os.path.exists(img_folder):
            os.mkdir(img_folder)

        if not categories:
            categories = []
            for c_name in dataset.labels:
                categories.append({'id': len(categories) + 1, 'name': c_name})

        logger.info(f'start conversion for {args.name}...')
        ic_dataset = DetectionAsClassificationDataset(dataset, aug_params)

        for img, labels, idx in ic_dataset:
            img_id = int(idx) + 1
            file_name = f'{idx}.{img.format}'
            img.save(os.path.join(img_folder, file_name), img.format)
            logger.log(logging.DEBUG, f'Saving to {os.path.join(img_folder, file_name)}')
            file_name = f'{phase}/{file_name}'
            images.append({'id': img_id, 'file_name': file_name, 'width': img.width, 'height': img.height})
            annotations.append({'id': len(annotations) + 1, 'image_id': img_id, 'category_id': labels[0] + 1})

        with open(f'{args.output_folder}/{phase}.json', 'w') as coco_out:
            coco_out.write(json.dumps({'images': images, 'categories': categories, 'annotations': annotations}, indent=2))


if __name__ == '__main__':
    main()



"""Download a dataset from shared storage"""
import argparse
import logging
import os
import shutil

from vision_datasets.common.dataset_downloader import DatasetDownloader
from vision_datasets.common.dataset_registry import DatasetRegistry

logger = logging.getLogger()


def download(dataset_sas_url: str, dataset_name: str, output_dir):
    assert dataset_sas_url
    assert dataset_name
    assert output_dir

    if os.path.exists(output_dir):
        raise RuntimeError(f'Output directory already exists: {output_dir}')

    dataset_registry = DatasetRegistry()
    downloader = DatasetDownloader(dataset_sas_url, dataset_registry)
    with downloader.download(dataset_name) as downloaded:
        for x in downloaded.base_dirs:
            shutil.copytree(x, output_dir)

    logger.info(f'Downloaded {dataset_name} to {output_dir}')


def list_datasets():
    registry = DatasetRegistry()
    for dataset in registry.list_data_version_and_types():
        logger.info(f"Name: {dataset['name']}, version: {dataset['version']}, type: {dataset['type']}")


def convert_to_tsv(dataset, file_path, idx_prefix):
    import json
    from io import BytesIO
    import base64
    from tqdm import tqdm

    idx = 0
    with open(file_path, 'w') as file_out:
        for sample in tqdm(dataset, desc=f'Writing to {file_path}'):
            img_id = f'{idx_prefix}{idx}'
            labels = []
            for label in sample[1]:
                tag_name = dataset.labels[label[0]]
                rect = label[1:5]
                labels.append({'class': tag_name, 'rect': rect})

            buffered = BytesIO()
            sample[0].save(buffered, format='JPEG')
            sample[0].close()

            b64img = base64.b64encode(buffered.getvalue()).decode('utf-8')
            file_out.write(f'{img_id}\t{json.dumps(labels)}\t{b64img}\n')
            idx += 1


def main():
    parser = argparse.ArgumentParser('Download datasets from the shared storage')
    parser.add_argument('dataset_name', nargs='?',
                        help='Dataset name. If not specified, show a list of available datasets')
    parser.add_argument('--output', '-o', help='Output directory.', required=True)
    parser.add_argument('--dataset_sas_url', '-k', help='url to dataset folder.', required=True)
    parser.add_argument('--to_tsv', '-t', help='to tsv format or not.', type=bool, default=False)

    args = parser.parse_args()
    if args.dataset_name:
        if args.to_tsv:
            import tempfile
            import pathlib
            dataset_info = DatasetRegistry().get_dataset_info(args.dataset_name)

            temp_dir = tempfile.mkdtemp()
            try:
                dataset_info.root_folder = pathlib.Path(temp_dir) / pathlib.Path(dataset_info.root_folder)
                download(args.dataset_sas_url, args.dataset_name, dataset_info.root_folder)
                from vision_datasets.common.manifest_dataset import ManifestDataset
                from vision_datasets.common.constants import Usages
                for usage in [Usages.TRAIN_PURPOSE, Usages.TEST_PURPOSE]:
                    dataset = ManifestDataset(dataset_info, usage, coordinates='absolute')
                    if not os.path.exists(args.output):
                        os.mkdir(args.output)
                    convert_to_tsv(dataset, pathlib.Path(args.output) / f'{usage}.tsv', usage)
                    dataset.close()
            finally:
                shutil.rmtree(dataset_info.root_folder)
        else:
            download(args.dataset_sas_url, args.dataset_name, args.output)

    else:
        list_datasets()


if __name__ == '__main__':
    main()



import logging
import os.path
import pathlib
from copy import deepcopy
import random
from PIL import Image

from .base_dataset import BaseDataset
from .constants import DatasetTypes
from .dataset_info import BaseDatasetInfo
from .data_manifest import DatasetManifest
from .image_loader import PILImageLoader
from .util import FileReader

logger = logging.getLogger(__name__)


class ManifestDataset(BaseDataset):
    """Dataset class that accesses data from dataset manifest
    """

    def __init__(self, dataset_info: BaseDatasetInfo, dataset_manifest: DatasetManifest, coordinates='relative', dataset_resources=None):
        """

        Args:
            dataset_info (BaseDatasetInfo): dataset info, containing high level information about the dataset, such as name, type, description, etc
            dataset_manifest (DatasetManifest): dataset manifest containing meta data such as image paths, annotations, etc
            coordinates (str): 'relative' or 'absolute', indicating the desired format of the bboxes returned.
            dataset_resources (str): disposable resources associated with this dataset
        """
        assert dataset_manifest
        assert coordinates in ['relative', 'absolute']

        super().__init__(dataset_info)

        self.dataset_manifest = dataset_manifest
        self.coordinates = coordinates
        self._file_reader = FileReader()
        self.dataset_resources = dataset_resources

    @property
    def labels(self):
        return self.dataset_manifest.labelmap

    def __len__(self):
        return len(self.dataset_manifest.images)

    def _get_single_item(self, index):
        image_manifest = self.dataset_manifest.images[index]
        image = self._load_image(image_manifest.img_path)
        target = image_manifest.labels
        if self.coordinates == 'relative':
            w, h = image.size
            target = ManifestDataset._box_convert_to_relative(image_manifest.labels, w, h, self.dataset_info)

        return image, target, str(index)

    def close(self):
        self._file_reader.close()

    def _load_image(self, filepath):
        full_path = filepath.replace('\\', '/')
        try:
            with self._file_reader.open(full_path, 'rb') as f:
                return PILImageLoader.load_from_stream(f)
        except Exception:
            logger.exception(f'Failed to load an image with path: {full_path}')
            raise

    @staticmethod
    def _box_convert_to_relative(target, w, h, dataset_info):
        # Convert absolute coordinates to relative coordinates.
        # Example: for image with size (200, 200), (1, 100, 100, 200, 200) => (1, 0.5, 0.5, 1.0, 1.0)
        if dataset_info.type == DatasetTypes.MULTITASK:
            return {task_name: ManifestDataset._box_convert_to_relative(task_target, w, h, dataset_info.sub_task_infos[task_name]) for task_name, task_target in target.items()}
        if dataset_info.type == DatasetTypes.OD:
            return [[t[0], t[1] / w, t[2] / h, t[3] / w, t[4] / h] for t in target]

        return target


class DetectionAsClassificationDataset(BaseDataset):
    """
    Consume detection dataset as a classification dataset, i.e., sample from this dataset is a crop wrt a bbox in the detection dataset.
    """

    def __init__(self, detection_dataset: ManifestDataset, box_aug_params: dict = None, local_cache_params: dict = None):
        """
        Args:
            detection_dataset: the detection dataset where images are cropped as classification samples
            box_aug_params (dict): params controlling box crop augmentation,
                'zoom_ratio_bounds': the lower/upper bound of box zoom ratio wrt box width and height, e.g., (0.3, 1.5)
                'shift_relative_bounds': lower/upper bounds of relative ratio wrt box width and height that a box can shift, e.g., (-0.3, 0.1)
                'rnd_seed': rnd seed used for box crop zoom and shift
            local_cache_params(dict): params controlling local cache for crop access:
                'dir': local dir for caching crops, it will be auto-created if not exist
                'n_max_copies': max number of crops cached for each bbox
        """
        assert detection_dataset
        assert detection_dataset.dataset_info.type == DatasetTypes.OD

        dataset_info = deepcopy(detection_dataset.dataset_info)
        dataset_info.type = DatasetTypes.IC_MULTICLASS
        super().__init__(dataset_info)

        self._dataset = detection_dataset
        self._n_booxes = 0
        self._box_abs_id_to_img_rel_id = {}
        for img_id, x in enumerate(self._dataset):
            boxes = x[1]
            for i in range(len(boxes)):
                self._box_abs_id_to_img_rel_id[self._n_booxes] = (img_id, i)
                self._n_booxes += 1
        self._box_aug_params = box_aug_params

        self._box_aug_rnd = random.Random(self._box_aug_params['rnd_seed']) if box_aug_params else None
        self._box_pick_rnd = random.Random(0)
        self._local_cache_params = local_cache_params
        if self._local_cache_params and not os.path.exists(self._local_cache_params['dir']):
            os.makedirs(self._local_cache_params['dir'])

    @property
    def labels(self):
        return self._dataset.labels

    def __len__(self):
        return self._n_booxes

    def _get_single_item(self, index):
        local_img_path = None
        img_idx, box_rel_idx = self._box_abs_id_to_img_rel_id[index]
        if self._local_cache_params:
            box_copy_idx = self._box_pick_rnd.randint(0, self._local_cache_params["max_n_copies"] - 1)
            box_img_id = f'{index}-{box_copy_idx}' if self._box_aug_params else str(index)
            local_img_path = pathlib.Path(self._local_cache_params['dir']) / box_img_id
            if os.path.exists(local_img_path):
                logger.log(logging.DEBUG, f'Found local cache for crop {index}! {box_copy_idx}')
                c_id = self._dataset.dataset_manifest.images[img_idx].labels[box_rel_idx][0]
                return Image.open(local_img_path), [c_id], str(index)

        img, boxes, _ = self._dataset[img_idx]
        c_id, left, t, r, b = boxes[box_rel_idx]
        if self._dataset.coordinates == 'relative':
            w, h = img.size
            left *= w
            t *= h
            r *= w
            b *= h

        box_img = DetectionAsClassificationDataset.crop(img, left, t, r, b, self._box_aug_params, self._box_aug_rnd)
        if local_img_path:
            box_img.save(local_img_path, box_img.format)
        return box_img, [c_id], str(index)

    def close(self):
        self._dataset.close()

    @staticmethod
    def crop(img, left, t, r, b, aug_params=None, rnd: random.Random = None):
        if aug_params:
            assert rnd
            if 'zoom_ratio_bounds' in aug_params:
                ratio_lower_b, ratio_upper_b = aug_params['zoom_ratio_bounds']
                left, t, r, b = BoxAlteration.zoom_box(left, t, r, b, img.size[0], img.size[1], ratio_lower_b, ratio_upper_b, rnd)

            if 'shift_relative_bounds' in aug_params:
                relative_lower_b, relative_upper_b = aug_params['shift_relative_bounds']
                left, t, r, b = BoxAlteration.shift_box(left, t, r, b, img.size[0], img.size[1], relative_lower_b, relative_upper_b, rnd)

        crop_img = img.crop((left, t, r, b))
        crop_img.format = img.format

        return crop_img


class BoxAlteration:
    @staticmethod
    def _stay_in_range(val, low, up):
        return int(min(max(val, low), up))

    @staticmethod
    def shift_box(left, t, r, b, img_w, img_h, relative_lower_b, relative_upper_b, rnd: random.Random):
        level = logging.DEBUG
        logger.log(level, f'old box {left}, {t}, {r}, {b}, out of ({img_w}, {img_h})')
        box_w = r - left
        box_h = b - t
        hor_shift = rnd.uniform(relative_lower_b, relative_upper_b) * box_w
        ver_shift = rnd.uniform(relative_lower_b, relative_upper_b) * box_h
        left = BoxAlteration._stay_in_range(left + hor_shift, 0, img_w)
        t = BoxAlteration._stay_in_range(t + ver_shift, 0, img_h)
        r = BoxAlteration._stay_in_range(r + hor_shift, 0, img_w)
        b = BoxAlteration._stay_in_range(b + ver_shift, 0, img_h)
        logger.log(level, f'[shift_box] new box {left}, {t}, {r}, {b}, with {hor_shift}, {ver_shift}, out of ({img_w}, {img_h})')

        return left, t, r, b

    @staticmethod
    def zoom_box(left, t, r, b, img_w, img_h, ratio_lower_b, ratio_upper_b, rnd: random.Random):
        level = logging.DEBUG
        logger.log(level, f'old box {left}, {t}, {r}, {b}, out of ({img_w}, {img_h})')
        w_ratio = rnd.uniform(ratio_lower_b, ratio_upper_b)
        h_ratio = rnd.uniform(ratio_lower_b, ratio_upper_b)
        box_w = r - left
        box_h = b - t
        new_box_w = box_w * w_ratio
        new_box_h = box_h * h_ratio
        logger.log(level, f'w h change: {box_w} {box_h} => {new_box_w} {new_box_h}')
        left = BoxAlteration._stay_in_range(left - (new_box_w - box_w) / 2, 0, img_w)
        t = BoxAlteration._stay_in_range(t - (new_box_h - box_h) / 2, 0, img_h)
        r = BoxAlteration._stay_in_range(r + (new_box_w - box_w) / 2, left, img_w)
        b = BoxAlteration._stay_in_range(b + (new_box_h - box_h) / 2, 0, img_h)
        logger.log(level, f'[zoom_box] new box {left}, {t}, {r}, {b}, with {w_ratio}, {h_ratio}, out of ({img_w}, {img_h})')

        return left, t, r, b



from .manifest_dataset import ManifestDataset
from .dataset_registry import DatasetRegistry
from .dataset_info import BaseDatasetInfo, DatasetInfo, MultiTaskDatasetInfo
from .data_manifest import DatasetManifest, CocoManifestAdaptor, IrisManifestAdaptor
from .constants import Usages, DatasetTypes

__all__ = ['ManifestDataset', 'DatasetRegistry', 'BaseDatasetInfo', 'DatasetInfo', 'MultiTaskDatasetInfo', 'Usages', 'DatasetTypes', 'DatasetManifest', 'CocoManifestAdaptor', 'IrisManifestAdaptor']



import argparse
import logging
import os.path
import pathlib
import random

from vision_datasets import DatasetRegistry, Usages, DatasetHub

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def show_dataset_stats(dataset):
    logger.info(f'Dataset stats: #images {len(dataset)}, #tags {len(dataset.labels)}')


def show_img(sample):
    sample[0].show()
    sample[0].close()

    logger.info(f'label = {sample[1]}')


def logging_prefix(dataset_name, version):
    return f'Dataset check {dataset_name}, version {version}: '


def check_dataset(dataset):
    show_dataset_stats(dataset)
    for idx in random.sample(range(len(dataset)), min(10, len(dataset))):
        show_img(dataset[idx])


def main():
    parser = argparse.ArgumentParser('Check if a dataset is valid')
    parser.add_argument('name', type=str, help="Dataset name.")
    parser.add_argument('--reg_json', '-r', type=str, default=None, help="dataset registration json file path.", required=True)
    parser.add_argument('--version', '-v', type=int, help="Dataset version.", default=None)
    parser.add_argument('--blob_container', '-k', type=str, help="blob container (sas) url", required=False)
    parser.add_argument('--folder_to_check', '-f', type=str, required=False, help="Check the dataset in this folder.")

    args = parser.parse_args()
    prefix = logging_prefix(args.name, args.version)

    data_reg_json = pathlib.Path(args.reg_json).read_text()
    dataset_info = DatasetRegistry(data_reg_json).get_dataset_info(args.name, args.version)
    if not dataset_info:
        logger.error(f'{prefix} dataset does not exist.')
        return
    else:
        logger.info(f'{prefix} dataset found in registration file.')

    vision_datasets = DatasetHub(data_reg_json)

    for usage in [Usages.TRAIN_PURPOSE, Usages.VAL_PURPOSE, Usages.TEST_PURPOSE]:
        logger.info(f'{prefix} Check dataset with usage: {usage}.')
        if args.folder_to_check and not os.path.exists(args.folder_to_check):
            os.mkdir(args.folder_to_check)

        # if args.folder_to_check is none, then this check will directly try to access data from azure blob. Images must be present in uncompressed folder on azure blob.
        dataset = vision_datasets.create_manifest_dataset(container_sas=args.blob_container, local_dir=args.folder_to_check, name=dataset_info.name, version=args.version, usage=usage)
        if dataset:
            check_dataset(dataset)
        else:
            logger.info(f'{prefix} No split for {usage} available.')


if __name__ == '__main__':
    main()



import setuptools
from os import path

VERSION = '0.2.4'

# Get the long description from the README file
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), 'r') as f:
    long_description = f.read()

setuptools.setup(name='vision_datasets',
                 author='Ping Jin, Shohei Ono',
                 description='A utility repo for vision dataset access and management.',
                 long_description=long_description,
                 long_description_content_type='text/markdown',
                 url='https://github.com/microsoft/vision-datasets',
                 version=VERSION,
                 python_requires='>=3.6',
                 license='MIT',
                 keywords='vision datasets classification detection',
                 packages=setuptools.find_packages(),
                 package_data={'': ['resources/*']},
                 install_requires=[
                     'numpy>=1.18.3',
                     'Pillow>=6.2.2',
                     'requests>=2.23.0',
                     'tenacity>=6.2.0'
                 ],
                 classifiers=[
                     'Development Status :: 4 - Beta',
                     'Intended Audience :: Developers',
                     'License :: OSI Approved :: MIT License',
                     'Programming Language :: Python :: 3.6',
                     'Programming Language :: Python :: 3.7',
                     'Programming Language :: Python :: 3.8',
                     'Programming Language :: Python :: 3.9',
                 ],
                 extras_require={'run': ['torch>=1.6.0', 'torchvision>=0.7.0']},
                 entry_points={
                     'console_scripts': ['vision_download=vision_datasets.commands.download:main',
                                         'vision_check_dataset=vision_datasets.commands.dataset_check:main',
                                         'vision_od_to_ic=vision_datasets.commands.converter_od_to_ic:main',
                                         'vision_convert_local_dir_ic_data=vision_datasets.commands.convert_local_dir_ic_data:main',
                                         'vision_tsv_to_iris=vision_datasets.commands.converter_tsv_to_iris:main',
                                         'vision_tsv_to_coco=vision_datasets.commands.converter_tsv_to_coco:main',
                                         'vision_yolo_to_iris=vision_datasets.commands.converter_yolo_darknet_to_iris:main',
                                         'vision_gen_metafile=vision_datasets.commands.generate_image_meta_info:main']
                 })



import copy
import json
from .dataset_info import DatasetInfoFactory
from typing import Union


class DatasetRegistry:
    """
    A central registry of all available datasets
    """

    def __init__(self, datasets_json: Union[str, list]):
        if isinstance(datasets_json, list):
            self.datasets = [DatasetInfoFactory.create(d) for dj in datasets_json for d in json.loads(dj)]
        else:
            self.datasets = [DatasetInfoFactory.create(d) for d in json.loads(datasets_json)]

    def get_dataset_info(self, dataset_name, dataset_version=None):
        datasets = [d for d in self.datasets if d.name == dataset_name and (not dataset_version or d.version == dataset_version)]
        if not datasets:
            return None

        sorted_datasets = sorted(datasets, key=lambda d: d.version)
        return copy.deepcopy(sorted_datasets[-1])

    def list_data_version_and_types(self):
        return [{'name': d.name, 'version': d.version, 'type': d.type, 'description': d.description} for d in self.datasets]

    @staticmethod
    def _get_default_dataset_json(json_file_name):
        import sys
        py_version = sys.version_info
        if py_version.minor >= 7:
            import importlib.resources as pkg_resources
            from vision_datasets import resources
            datasets_json = pkg_resources.read_text(resources, json_file_name)
        else:
            import pkgutil
            resource_package = 'vision_datasets'
            resource_path = '/'.join(('resources', json_file_name))
            datasets_json = pkgutil.get_data(resource_package, resource_path)
        return datasets_json
